from textwrap import dedent
from mccode_antlr.loader import parse_mcstas_instr
from mccode_antlr.test import compiled_test
from mccode_antlr.utils import compile_and_run

@compiled_test
def test_when_logical_parses():
    instr = parse_mcstas_instr(dedent("""DEFINE INSTRUMENT when_logical(int should_eye=0)
    DECLARE %{int copied;%}
    INITIALIZE %{copied = should_eye;%}
    TRACE
    COMPONENT only = Arm() WHEN(should_eye) at (0, 0, 0) ABSOLUTE
    EXTEND %{
    printf("should_eye=%d\\n", _instrument_var._parameters.should_eye);
    %}
    COMPONENT acopy = Arm() WHEN(copied) at (0, 0, 1) ABSOLUTE
    EXTEND %{
    printf("copied=%d\\n", copied);
    %} 
    END
    """))
    for should_eye in (0, 1):
        results, data = compile_and_run(instr, f'-n 1 {should_eye=}')
        lines = results.decode('utf-8').splitlines()
        expected = [f'{should_eye=}', f'copied={should_eye}'] if should_eye else []
        assert len(expected) + 1 == len(lines)  # [*expected, '*** TRACE end *** ']
        for ex, ln in zip(expected, lines):
            assert ex == ln


def do_when_op_(checker, op, message):
    from pathlib import Path
    from tempfile import TemporaryDirectory
    from mccode_antlr import Flavor
    from mccode_antlr.run import mccode_compile, mccode_run_compiled
    from itertools import product

    kwargs = dict(target=None, config=None, dump_source=False)

    instr = parse_mcstas_instr(
        dedent(f"""DEFINE INSTRUMENT when_equal(int value=0, int check=-1)
        initialize %{{
        printf("value=%d check=%d\\n", value, check);
        %}}
        TRACE
        COMPONENT only = Arm() WHEN(value {op} check) at (0, 0, 0) ABSOLUTE
        EXTEND %{{
        printf("{message}\\n");
        %}}
        END
        """)
    )
    with TemporaryDirectory() as directory:
        binary, target = mccode_compile(instr, directory, Flavor.MCSTAS, **kwargs)
        for index, (value, check) in enumerate(product((0, 1, 2, 3), (1, 2, 3, 4))):
            path = Path(directory).joinpath(f't{index}')
            result, data = mccode_run_compiled(binary, target, path, f'-n 1 {value=} {check=}')
            lines = result.decode('utf-8').splitlines()
            # print(f'{value=} {check=} {op=} {lines=}')
            expected = [f'{value=} {check=}'] + ([message] if checker(value, check) else [])
            assert len(expected) + 1 == len(lines)  # [*expected, '*** TRACE end *** ']
            for ex, ln in zip(expected, lines):
                assert ex == ln


def test_when_equal_parses():
    do_when_op_(lambda x, y: x == y, '==', 'equal')


def test_when_not_equal_parses():
    do_when_op_(lambda x, y: x != y, '!=', 'unequal')




from mccode_antlr.io.msgpack import to_msgpack, from_msgpack

def test_simple_instr_json():
    from mccode_antlr.loader import parse_mcstas_instr
    instr = parse_mcstas_instr(
        "define instrument check() trace component a = Arm() at (0,0,0) absolute end")
    msg = to_msgpack(instr)
    reconstituted = from_msgpack(msg)
    assert type(reconstituted) is type(instr)
    assert instr == reconstituted




from textwrap import dedent
from mccode_antlr.loader import parse_mcstas_instr
from mccode_antlr.test import compiled_test
from mccode_antlr.utils import compile_and_run
from mccode_antlr.reader.registry import InMemoryRegistry


FAKE_COMPONENTS = dict(
    n_part=dedent("""DEFINE COMPONENT n_part
    SETTING PARAMETERS (int n)
    TRACE
    %{
      for (int i= 0; i < n; i++) {
        printf("n=%d\\n", i);
      }
    %}
    END
    """),
    m_part=dedent("""DEFINE COMPONENT m_part
    SETTING PARAMETERS (int m)
    TRACE
    %{
      for (int i = 0; i < m; i++) {
        printf("m=%d\\n", i);
      }
    %}
    END
    """),
    both_parts=dedent("""DEFINE COMPONENT both_parts
    SETTING PARAMETERS (int n, int m)
    TRACE INHERIT n_part INHERIT m_part
    END
    """),
    crazy=dedent("""DEFINE COMPONENT crazy
    SETTING PARAMETERS (int n, int m, int k)
    TRACE %{
    printf("We can do anything\\n");
    %}
    inherit n_part 
    EXTEND %{
    printf("Why!?\\n"); 
    %}
    inherit m_part
    EXTEND %{
    printf("Because we can!\\n"); 
    %}
    END"""),
)


in_memory = InMemoryRegistry("test_components")
for comp, repr in FAKE_COMPONENTS.items():
    in_memory.add_comp(comp, repr)


@compiled_test
def test_n_part():
    instr = parse_mcstas_instr(dedent("""
    define instrument test_n_part(dummy=0.)
    trace
    component origin = n_part(n=1) at (0, 0, 0) absolute
    end
    """), registries=[in_memory])
    output, files = compile_and_run(instr, '-n 1 dummy=2')
    lines = output.decode('utf-8').splitlines()
    for line, expected in zip(lines, ("n=0",)):
        assert line.strip() == expected.strip()


@compiled_test
def test_m_part():
    instr = parse_mcstas_instr(dedent("""
    define instrument test_m_part(dummy=0.)
    trace
    component origin = m_part(m=2) at (0, 0, 0) absolute
    end
    """), registries=[in_memory])
    output, files = compile_and_run(instr, '-n 1 dummy=2')
    lines = output.decode('utf-8').splitlines()
    for line, expected in zip(lines, ("m=0", "m=1")):
        assert line.strip() == expected.strip()


@compiled_test
def test_both_parts():
    instr = parse_mcstas_instr(dedent("""
    define instrument test_both_parts(dummy=0.)
    trace
    component origin = both_parts(n=3, m=2) at (0, 0, 0) absolute
    end
    """), registries=[in_memory])
    output, files = compile_and_run(instr, '-n 1 dummy=2')
    lines = output.decode('utf-8').splitlines()
    for line, expected in zip(lines, ("n=0", "n=1", "n=2", "m=0", "m=1")):
        assert line.strip() == expected.strip()


@compiled_test
def test_crazy_parts():
    instr = parse_mcstas_instr(dedent("""
    define instrument test_crazy(dummy=0.)
    trace
    component origin = crazy(n=3, m=2, k=0) at (0, 0, 0) absolute
    end
    """), registries=[in_memory])
    output, files = compile_and_run(instr, '-n 1 dummy=2')
    lines = output.decode('utf-8').splitlines()
    for line, expected in zip(lines, ("We can do anything", "n=0", "n=1", "n=2", "Why!?", "m=0", "m=1", "Because we can!")):
        assert line.strip() == expected.strip()


from unittest import TestCase
from pathlib import Path
from mccode_antlr import Flavor
import mccode_antlr.config

MCKEY = 'MCCODEANTLR_MCSTAS__PATHS'
MXKEY = 'MCCODEANTLR_MCXTRACE__PATHS'


class TestLocalRegistryCollection(TestCase):
    def setUp(self):
        from tempfile import mkdtemp
        self.temps = [Path(p) for p in [mkdtemp(), mkdtemp()]]

    def tearDown(self):
        from shutil import rmtree
        for tmp in self.temps:
            rmtree(tmp)

    def assertAllEqual(self, a, b):
        self.assertEqual(len(a), len(b))
        for i, j in zip(a, b):
            self.assertEqual(i, j)

    def assertPaths(self, flavor, paths: list[Path]):
        from importlib import reload
        reload(mccode_antlr.config)
        from mccode_antlr.reader import LocalRegistry as LReg
        from mccode_antlr.reader.registry import REGISTRY_PRIORITY_HIGH as HIGH
        from mccode_antlr.reader.registry import default_registries
        def_regs = default_registries(flavor)
        for x in [LReg(path.stem, path.as_posix(), priority=HIGH) for path in paths]:
            self.assertIn(x, def_regs)

    def test_mcstas_environment_variable_single(self):
        import os
        from unittest.mock import patch
        with patch.dict(os.environ, {MCKEY: self.temps[0].as_posix()}):
            self.assertPaths(Flavor.MCSTAS, self.temps[0:1])

    def test_mcxtrace_environment_variable_single(self):
        import os
        from unittest.mock import patch
        with patch.dict(os.environ, {MXKEY: self.temps[1].as_posix()}):
            self.assertPaths(Flavor.MCXTRACE, self.temps[1:2])

    def test_mcstas_environment_variable_multi(self):
        import os
        from unittest.mock import patch
        with patch.dict(os.environ, {MCKEY: ' '.join(p.as_posix() for p in self.temps)}):
            self.assertPaths(Flavor.MCSTAS, self.temps[0:2])

    def test_mcxtrace_environment_variable_multi(self):
        import os
        from unittest.mock import patch
        with patch.dict(os.environ, {MXKEY: ' '.join(p.as_posix() for p in self.temps)}):
            self.assertPaths(Flavor.MCXTRACE, self.temps[0:2])


from textwrap import dedent
from mccode_antlr.loader import parse_mcstas_instr, parse_mcxtrace_instr
from mccode_antlr.test import compiled_test
from mccode_antlr.utils import compile_and_run, Flavor
from mccode_antlr.reader.registry import InMemoryRegistry

in_memory = InMemoryRegistry(
    "test_components",
    calls_restore_xray=dedent("""DEFINE COMPONENT calls_restore_xray
    SETTING PARAMETERS (int n)
    TRACE %{
    RESTORE_XRAY(INDEX_CURRENT_COMP, x, y, z, kx, ky, kz, phi, t, Ex, Ey, Ez, p);
    %}
    END
    """),
    calls_restore_neutron=dedent("""DEFINE COMPONENT calls_restore_neutron
    SETTING PARAMETERS (int n)
    TRACE %{
    RESTORE_NEUTRON(INDEX_CURRENT_COMP, x, y, z, vx, vy, vz, t, sx, sy, sz, p);
    %}
    END
    """),
    useless=dedent("""DEFINE COMPONENT useless
    SETTING PARAMETERS (int n)
    TRACE 
    %{
    %}
    END
    """),
)


@compiled_test
def test_simplest_mcstas():
    instr = parse_mcstas_instr(dedent("""
    define instrument test_simplest_mcstas(dummy=0.)
    trace component origin = useless() at (0, 0, 0) absolute end
    """), registries=[in_memory])
    compile_and_run(instr, '-n 1 dummy=2', flavor=Flavor.MCSTAS)


@compiled_test
def test_simplest_mcxtrace():
    instr = parse_mcxtrace_instr(dedent("""
    define instrument test_simplest_mcxtrace(dummy=0.)
    trace component origin = useless() at (0, 0, 0) absolute end
    """), registries=[in_memory])
    compile_and_run(instr, '-n 1 dummy=2', flavor=Flavor.MCXTRACE)


@compiled_test
def test_restore_neutron():
    instr = parse_mcstas_instr(dedent("""
    define instrument test_restore_neutron(dummy=0.)
    trace component origin = calls_restore_neutron() at (0, 0, 0) absolute end
    """), registries=[in_memory])
    compile_and_run(instr, '-n 1 dummy=2', flavor=Flavor.MCSTAS)

@compiled_test
def test_restore_xray():
    instr = parse_mcxtrace_instr(dedent("""
    define instrument test_restore_xray(dummy=0.)
    trace component origin = calls_restore_xray() at (0, 0, 0) absolute end
    """), registries=[in_memory])
    compile_and_run(instr, '-n 1 dummy=2', flavor=Flavor.MCXTRACE)


#!/usr/bin/env python3
"""
Test to reproduce the nested key save issue.
"""
import tempfile
from pathlib import Path
import yaml


def test_nested_key_save_issue():
    """Reproduce the issue where setting a nested key only saves that key."""
    from mccode_antlr.cli.management import config_set

    with tempfile.TemporaryDirectory() as tmpdir:
        tmppath = Path(tmpdir)

        # Manually create an initial config with nested structure
        config_file = tmppath / 'config.yaml'
        initial_config = {
            'compiler': {
                'cc': 'gcc',
                'flags': '-O2',
                'debug': False
            },
            'runtime': {
                'threads': 4
            }
        }
        with config_file.open('w') as f:
            yaml.dump(initial_config, f)

        print("Initial config:")
        print(yaml.dump(initial_config))

        # Now use config_set to update a nested key
        config_set('compiler.flags', '-O3', str(tmppath))

        # Read back the saved config
        with config_file.open('r') as f:
            saved_config = yaml.safe_load(f)

        print("\nSaved config after setting compiler.flags to -O3:")
        print(yaml.dump(saved_config))

        # Check if all keys are preserved
        print("\nChecking preserved keys:")
        print(f"  compiler.cc: {'✓' if 'cc' in saved_config.get('compiler', {}) else '✗ MISSING'}")
        print(f"  compiler.flags: {'✓' if 'flags' in saved_config.get('compiler', {}) else '✗ MISSING'}")
        print(f"  compiler.debug: {'✓' if 'debug' in saved_config.get('compiler', {}) else '✗ MISSING'}")
        print(f"  runtime.threads: {'✓' if 'threads' in saved_config.get('runtime', {}) else '✗ MISSING'}")

        # Verify the issue
        assert 'compiler' in saved_config, "compiler section missing!"
        assert 'runtime' in saved_config, "runtime section missing!"
        assert saved_config['compiler']['flags'] == '-O3', "flags not updated!"
        assert saved_config['compiler']['cc'] == 'gcc', "cc was lost!"
        assert saved_config['runtime']['threads'] == 4, "runtime section was lost!"


if __name__ == '__main__':
    test_nested_key_save_issue()
    print("\n✅ Test passed!")




from __future__ import annotations
from functools import cache
import pytest

from mccode_antlr.compiler.check import simple_instr_compiles


def compiled_test(method, compiler: str | None = None):
    if compiler is None:
        compiler = 'cc'
    if simple_instr_compiles(compiler):
        return method

    @pytest.mark.skip(reason=f"Skipping due to lack of working {compiler}")
    def skipped_method(*args, **kwargs):
        return method(*args, **kwargs)

    return skipped_method


def gpu_compiled_test(method):
    return compiled_test(method, 'acc')


def mpi_compiled_test(method):
    return compiled_test(method, 'mpi/cc')


@cache
def mcpl_config_available():
    from shutil import which
    return which('mcpl-config') is not None


def mcpl_compiled_test(method):
    if not mcpl_config_available():
        @pytest.mark.skip(reason='mcpl-config not available')
        def no_mcpl(*args, **kwargs):
            return method(*args, **kwargs)
        return no_mcpl
    if not simple_instr_compiles('cc'):
        @pytest.mark.skip(reason='no working compiler cc')
        def no_cc(*args, **kwargs):
            return method(*args, **kwargs)
    return method




from .reader import Reader
from .registry import (
    Registry,
    LocalRegistry, RemoteRegistry, ModuleRemoteRegistry, GitHubRegistry,
    InMemoryRegistry,
    collect_local_registries, default_registries, ensure_registries
)

__all__ = [
    'Reader',
    'Registry',
    'LocalRegistry',
    'RemoteRegistry',
    'ModuleRemoteRegistry',
    'GitHubRegistry',
    'InMemoryRegistry',
    'collect_local_registries',
    'default_registries',
    'ensure_registries',
]



from __future__ import annotations

from mccode_antlr import Flavor

def run_prog_message_output(prog: list[str]):
    """Run a program and return its output

    Parameters
    ----------
    prog : list[str]
        The program to run, prog[0] should be the binary name, findable by shutil.which

    Returns
    -------
    message: str
        None if the program ran successfully, or a message indicating what went wrong
    output: str
        The standard output of the program
    """
    from shutil import which
    from subprocess import run
    message, output = None, None
    if which(prog[0]):
        res = run(prog, capture_output=True, text=True)
        output = res.stdout
        if res.returncode or len(res.stderr):
            message = f'Evaluating "{" ".join(prog)}" produced error: {res.stderr}'
    else:
        message = f'{prog[0]} not found'

    return message, output


def make_assembler(name: str, flavor: Flavor = Flavor.MCSTAS):
    from mccode_antlr.assembler import Assembler
    from mccode_antlr.reader.registry import default_registries
    return Assembler(name, registries=default_registries(flavor))


def parse_instr_string(instr_source: str):
    from mccode_antlr.loader import parse_mcstas_instr
    return parse_mcstas_instr(instr_source)



def compile_and_run(instr,
                    parameters,
                    run=True,
                    dump_source=True,
                    target: dict | None = None,
                    config: dict | None = None,
                    flavor: Flavor = Flavor.MCSTAS):
    from pathlib import Path
    from tempfile import TemporaryDirectory
    from mccode_antlr.run import mccode_compile, mccode_run_compiled

    kwargs = dict(target=target, config=config, dump_source=dump_source)

    with TemporaryDirectory() as directory:
        binary, target = mccode_compile(instr, directory, flavor=flavor, **kwargs)
        # The runtime output directory used *can not* exist for McStas/McXtrace to work properly.
        # So find a name inside this directory that doesn't exist (any name should work)
        return mccode_run_compiled(binary, target, Path(directory).joinpath('t'), parameters) if run else (None, None)



from textwrap import dedent
from .c_listener import CDeclarator
from ..instr import Instr
from mccode_antlr import Flavor


def header_pre_runtime(
        source: Instr,
        flavor: Flavor,
        config: dict,
        uservars: list[CDeclarator]
):
    from datetime import datetime
    from mccode_antlr import version
    from .c_particle import (
        accessible_struct_members, restorable_struct_members,
        setstate_signature_members, getstate_signature_members, setstate_signature_call
    )

    def jump_line(instance, jump):
        return f'long Jump_{instance.name}_{jump.target}; /* the JUMP connection <from>_<to> */'

    jump_string = '\n'.join([jump_line(i, j) for i in source.components for j in i.jump if j.iterate])

    if Flavor.MCSTAS == flavor:
        particle_struct = '\n'.join([
            "  double vx,vy,vz; /* velocity [m/s] */",
            "  double sx,sy,sz; /* spin [0-1] */",
            "  int mcgravitation; /* gravity-state */",
            "  void *mcMagnet;    /* precession-state */",
            "  int allow_backprop; /* allow backprop */"])
    elif Flavor.MCXTRACE == flavor:
        particle_struct = '\n'.join([
            "  double kx,ky,kz; /* wave-vector */",
            "  double phi, Ex,Ey,Ez; /* phase and electrical field */"])
    else:
        raise ValueError(f'Unknown flavor {flavor.name}')

    # Append variables from instr USERVARS block to particle struct
    # Also store these strings in the appropriate instrument list for later def/undef as state variables

    uservar_string = '//user variables and comp - injections:\n' if len(uservars) else ''
    uservar_string += '\n'.join([f'  {x};' for x in uservars])

    members = list(accessible_struct_members(flavor)) + [x.name for x in uservars]
    # Shouldn't we exclude array-valued names here?
    getvar = '\n'.join(
        [f'  if(!str_comp("{x}",name)){{rval=*((double*)(&(p->{x})));s=0;}}' for x in members]
    )

    # Array valued names seem OK here, since a user can type-cast correctly
    getvar_void = '\n'.join(
        [f'  if (!str_comp("{x}", name)){{rval=(void * ) & (p->{x});s=0;}}' for x in members]
    )

    # For non-array values we can safely use memcpy (probably)
    members_types = ([(x, y) for x, y in accessible_struct_members(flavor).items()]
                     + [(x.name, x.dtype) for x in uservars if not x.is_array and not x.is_pointer]
                     )
    setvar_void = '\n'.join(
        [f'  if(!str_comp("{x}",name)){{memcpy(&(p->{x}), value, sizeof({y})); rval=0;}}'
         for x, y in members_types]
    )

    # an array -- need to know how many elements we're copying
    setvar_void_array = '\n'.join(
        [f'  if(!str_comp("{x.name}",name)){{memcpy(&(p->{x.name}), value, elements * sizeof({x.dtype})); rval=0;}}'
         for x in uservars if x.is_array or x.is_pointer])

    restore = '\n'.join(
        [f'  p->{x} = p0->{x};' for x in restorable_struct_members(flavor)]
    )

    getuservar_byid = '\n'.join(
        [f'  case {i}: {{rval=*( (double *)(&(p->{x.name})) ); s=0; break;}}' for i, x in enumerate(uservars)])

    uservar_init = ''
    for x in uservars:
        if (x.dtype not in ('double', 'MCNUM', 'int') and x.init is None) or x.is_pointer or x.is_array:
            array_str = ' array' if x.is_pointer or x.is_array else ''
            print(f'\nWARNING:\n --> USERVAR {x.name} is of type {x.dtype}{array_str}')
            print('  --> and may need specific per-particle initialization through an EXTEND block!\n')
        else:
            uservar_init += f'\np->{x.name}={0 if x.init is None else x.init};'

    mcsetstate = ', '.join(f'{t} {n}' for n, t in setstate_signature_members(flavor).items())
    mcgetstate = ', '.join(f'{t} {n}' for n, t in getstate_signature_members(flavor).items())
    mcinitstate = ', '.join(setstate_signature_call(flavor))

    contents = dedent(f"""/* Automatically generated file. Do not edit.
     * Format:     ANSI C source code
     * Creator:    {flavor} <{flavor.url()}>
     * Generator:  mccode-antlr {version()} <https://github.com/McStasMcXtrace/mccode-antlr.git>
     * Instrument: {source.source} ({source.name})
     * Date: {datetime.now()}
     * File: {config.get('output')}
     * CFLAGS={' '.join(source.unique_flags)}
     */

    #ifndef WIN32
    #  ifndef OPENACC
    #    define _GNU_SOURCE
    #  endif
    #  define _POSIX_C_SOURCE 200809L
    #endif
    /* In case of cl.exe on Windows, suppress warnings about #pragma acc
       Transferred from https://github.com/McStasMcXtrace/McCode/commit/0e2785a2d3fd742d46597139234dbc47e56344bb 
    */
    #ifdef _MSC_EXTENSIONS
    #pragma warning(disable: 4068)
    #endif
     
    #define MCCODE_STRING "{flavor}"
    #define FLAVOR "{str(flavor).lower()}"
    #define FLAVOR_UPPER "{str(flavor).upper()}"
    {'#define MC_USE_DEFAULT_MAIN' if config.get('default_main') else ''}
    {'#define MC_TRACE_ENABLED' if config.get('enable_trace') else ''}
    {'#define MC_PORTABLE' if config.get('portable') else ''}
    
    #include <string.h>
    #include <inttypes.h>
    
    typedef double MCNUM;
    typedef struct {{MCNUM x, y, z;}} Coords;
    typedef MCNUM Rotation[3][3];
    #define MCCODE_BASE_TYPES
    
    /* available random number generators */
    #define _RNG_ALG_MT         1
    #define _RNG_ALG_KISS       2
    /* selection of random number generator */
    #ifndef RNG_ALG
    #  define RNG_ALG  _RNG_ALG_KISS
    #endif
    #if RNG_ALG == _RNG_ALG_MT // MT 
    #define randstate_t uint32_t
    #elif RNG_ALG == _RNG_ALG_KISS  // KISS
    #define randstate_t uint64_t
    #endif

    #ifndef MC_NUSERVAR
    #define MC_NUSERVAR 10
    #endif
    
    /* Particle JUMP control logic */
    struct particle_logic_struct {{
      int dummy;{jump_string}
    }};
    struct _struct_particle {{
      double x,y,z; /* position [m] */
    {particle_struct}
      /* Generic Temporaries: */
      /* May be used internally by components e.g. for special */
      /* return-values from functions used in trace, thusreturned via */
      /* particle struct. (Example: Wolter Conics from McStas, silicon slabs.) */
      double _mctmp_a; /* temp a */
      double _mctmp_b; /* temp b */
      double _mctmp_c; /* temp c */
      randstate_t randstate[7];
      double t, p;    /* time, event weight */
      long long _uid;  /* Unique event ID */
      long _index;     /* component index where to send this event */
      long _absorbed;  /* flag set to TRUE when this event is to be removed/ignored */
      long _scattered; /* flag set to TRUE when this event has interacted with the last component instance */
      long _restore;   /* set to true if neutron event must be restored */
      long flag_nocoordschange;   /* set to true if particle is jumping */
      struct particle_logic_struct _logic;
      {uservar_string}
    }};
    typedef struct _struct_particle _class_particle;
    
    _class_particle _particle_global_randnbuse_var;
    _class_particle* _particle = &_particle_global_randnbuse_var;
    
    // Below lines relating to mcgenstate / setstate are in principle McStas - centric, we ought to generate
    //this function based on "project"
    #pragma acc routine
    _class_particle mcgenstate(void);
    #pragma acc routine
    _class_particle mcsetstate({mcsetstate});
    #pragma acc routine
    _class_particle mcgetstate({mcgetstate});
    
    extern int mcgravitation;      /* flag to enable gravitation */
    #pragma acc declare create ( mcgravitation )
    int mcallowbackprop;        
    #pragma acc declare create ( mcallowbackprop )
    
    _class_particle mcgenstate(void) {{
      _class_particle particle = mcsetstate({mcinitstate});
      return(particle);
    }}
    /*Generated user variable handlers:*/
    
    #pragma acc routine
    double particle_getvar(_class_particle *p, char *name, int *suc);
    
    #ifdef OPENACC
    #pragma acc routine
    int str_comp(char *str1, char *str2);
    #endif
    
    double particle_getvar(_class_particle *p, char *name, int *suc){{
    #ifndef OPENACC
    #define str_comp strcmp
    #endif
      int s=1;
      double rval=0;
    {getvar}  
      if (suc!=0x0) {{*suc=s;}}
      return rval;
    }}
    
    #pragma acc routine
    void* particle_getvar_void(_class_particle *p, char *name, int *suc);\n
    #ifdef OPENACC
    #pragma acc routine
    int str_comp(char *str1, char *str2);
    #endif
    
    void* particle_getvar_void(_class_particle *p, char *name, int *suc){{
    #ifndef OPENACC
    #define str_comp strcmp
    #endif
      int s=1;
      void* rval=0;
    {getvar_void}
      if (suc!=0x0) {{*suc=s;}}
      return rval;
    }}
    
    #pragma acc routine
    int particle_setvar_void(_class_particle *, char *, void*);\n
    int particle_setvar_void(_class_particle *p, char *name, void* value){{
    #ifndef OPENACC
    #define str_comp strcmp
    #endif
      int rval=1;
    {setvar_void}  
      return rval;
    }}
    
    #pragma acc routine
    int particle_setvar_void_array(_class_particle *, char *, void*, int);
    
    int particle_setvar_void_array(_class_particle *p, char *name, void* value, int elements){{
    #ifndef OPENACC
    #define str_comp strcmp
    #endif
      int rval=1;
    {setvar_void_array}
      return rval;
    }}
    
    // Function to handle a particle restore of physical particle params
    #pragma acc routine
    void particle_restore(_class_particle *p, _class_particle *p0);
    void particle_restore(_class_particle *p, _class_particle *p0) {{
      {restore}
      p->_absorbed=0; p->_restore=0;
    }}
    
    #pragma acc routine
    double particle_getuservar_byid(_class_particle *p, int id, int *suc){{
      int s=1;
      double rval=0;
      switch(id){{
    {getuservar_byid}
      }}
      if (suc!=0x0) {{*suc=s;}}
      return rval;
    }}
    
    #pragma acc routine
    void particle_uservar_init(_class_particle *p){{
    {uservar_init}
    }}
    """)
    return contents


def header_post_runtime(source, flavor: Flavor, config: dict, include_path):
    from ..common.utilities import escape_str_for_c

    def source_file_contents():
        from pathlib import Path
        from os import access, R_OK
        if source.source is None and config.get('embed_instrument_file'):
            raise RuntimeError('Requested source embedding, but no source file provided!')
        elif source.source is None:
            return 'Instrument source code is not embedded in this executable.'
        path = Path(source.source)
        if config.get('embed_instrument_file') and path.exists() and access(path, R_OK):
            with path.open('r') as file:
                return escape_str_for_c(file.read())
        message = f"Instrument {source.name} source code "
        message += f"from {escape_str_for_c(source.source)} is not embedded in this executable.\\n"
        message += f"Use --source option when running {config.get('flavor')}"
        return message

    main_file_string = 'int main(int argc, char *argv[]){return mccode_main(argc, argv);}'
    contents = dedent(f"""
    /* *****************************************************************************
     * Start of instrument '{source.name}' generated code
    ***************************************************************************** */
    
    #ifdef MC_TRACE_ENABLED
    int traceenabled = 1;
    #else
    int traceenabled = 0;
    #endif
    #define {str(flavor).upper()} "{escape_str_for_c(str(include_path))}"
    int   defaultmain         = {1 if config.get("default_main") else 0};
    char  instrument_name[]   = "{source.name}";
    char  instrument_source[] = "{escape_str_for_c(source.source)}";
    char *instrument_exe      = NULL; /* will be set to argv[0] in main */
    char instrument_code[] = "{source_file_contents()}";
    {main_file_string if config.get('default_main') else ''}
    """)
    return contents



from mccode_antlr import Flavor


def restore_name(flavor: Flavor):
    particle = {Flavor.MCSTAS: "NEUTRON", Flavor.MCXTRACE: "XRAY"}[flavor]
    return f"RESTORE_{particle}"


def xyz_Axyz_Bxyz(flavor: Flavor, type_str):
    """Return lists of particle struct members

    Either:
        (x, y, z), (vx, vy, vz), (sx, sy, sz)
    or
        (x, y, z), (kx, ky, kz, phi), (Ex, Ey, Ez)
    """
    xyz = ('x', 'y', 'z')
    ps = {Flavor.MCSTAS: ('', 'v', 's'), Flavor.MCXTRACE: ('', 'k', 'E')}[flavor]
    z, a, b = (tuple((f'{p}{x}', type_str) for x in xyz) for p in ps)
    if Flavor.MCXTRACE == flavor:
        a = a + (('phi', type_str),)
    return z, a, b



def struct_members(flavor: Flavor):
    """Get the name and type of the members of the _class_particle struct

    The members of the struct are not equivalent between McStas and McXtrace
    """
    z, a, b = xyz_Axyz_Bxyz(flavor, 'double')
    if Flavor.MCSTAS == flavor:
        a = a +  (
            ('mcgravitation', 'int'),
            ('mcMagnet', 'void *'),
            ('allow_backprop', 'int')
        )

    members = (z + a + b) + (
        ('_mctmp_a', 'double'),
        ('_mctmp_b', 'double'),
        ('_mctmp_c', 'double'),
        ('randstate', 'unsigned long []'),
        ('t', 'double'),
        ('p', 'double'),
        ('_uid', 'long long'),
        ('_index', 'long'),
        ('_absorbed', 'long'),
        ('_scattered', 'long'),
        ('_restore', 'long'),
        ('flag_nocoordschange', 'long'),
        ('_logic', 'struct particle_logic_struct'),
        # uservars
    )
    return {k: v for k, v in members}

def accessible_struct_members(flavor: Flavor):
    """Get the name and type of the accessible members of the _class_particle struct

    The members of the struct are not equivalent between McStas and McXtrace

    Note
    ----
    Used in the generated C code by, e.g.,
        double particle_getvar(...)
        void particle_getvar_void(...)
        int particle_setvar_void(...)
        void particle_restore(...)
    written-out here within c_header.py
    """
    z, a, b = xyz_Axyz_Bxyz(flavor, 'double')
    members = (z + a + b) + (
        ('t', 'double'),
        ('p', 'double'),
        ('_mctmp_a', 'double'),
        ('_mctmp_b', 'double'),
        ('_mctmp_c', 'double'),
        # uservars
    )
    return {k: v for k, v in members}

def restorable_struct_members(flavor: Flavor):
    """Get the name and type of the restorable members of the _class_particle struct

    The members of the struct are not equivalent between McStas and McXtrace

    Note
    ----
    Used in the generated C code by, e.g.,
        void particle_restore(...)
    written-out here within c_header.py
    """
    z, a, b = xyz_Axyz_Bxyz(flavor, 'double')
    members = z + a + b + (('t', 'double'), ('p', 'double'),)
    return {k: v for k, v in members}


def setstate_signature_members(flavor: Flavor):
    z, a, b = xyz_Axyz_Bxyz(flavor, 'double')
    members = (
        z + a + (('t', 'double'),) + b
        + (
            ('p', 'double'),
            ('mcgravitation', 'int'),
            ('mcMagnet', 'void *'),
            ('allow_backprop', 'int')
        )
    )
    return {k: v for k, v in members}


def getstate_signature_members(flavor: Flavor):
    z, a, b = xyz_Axyz_Bxyz(flavor, 'double *')
    members = (
            (('mcparticle', '_class_particle'),)
            + z + a + (('t', 'double *'),) + b + (('p', 'double *'),)
    )
    return {k: v for k, v in members}



def setstate_signature_call(flavor: Flavor):
    values = {
        {Flavor.MCSTAS: 'vz', Flavor.MCXTRACE: 'kz'}[flavor]: '1',
        'p': '1',
        'mcgravitation': 'mcgravitation',
        'mcMagnet': 'NULL',
        'allow_backprop': 'mcallowbackprop',
    }
    members = setstate_signature_members(flavor)
    return [values.get(key, '0') for key in members]



#!/usr/bin/env python3
"""
Test to verify config_unset preserves other keys.
"""
import tempfile
from pathlib import Path
import yaml


def test_config_unset_preserves_keys():
    """Verify that config_unset only removes the target key and preserves others."""
    from mccode_antlr.cli.management import config_unset

    with tempfile.TemporaryDirectory() as tmpdir:
        tmppath = Path(tmpdir)

        # Manually create an initial config with nested structure
        config_file = tmppath / 'config.yaml'
        initial_config = {
            'compiler': {
                'cc': 'gcc',
                'flags': '-O2',
                'debug': False
            },
            'runtime': {
                'threads': 4,
                'verbose': True
            }
        }
        with config_file.open('w') as f:
            yaml.dump(initial_config, f)

        print("Initial config:")
        print(yaml.dump(initial_config))

        # Now use config_unset to remove a nested key
        config_unset('compiler.debug', str(tmppath))

        # Read back the saved config
        with config_file.open('r') as f:
            saved_config = yaml.safe_load(f)

        print("\nSaved config after unsetting compiler.debug:")
        print(yaml.dump(saved_config))

        # Check if other keys are preserved
        print("\nChecking preserved keys:")
        print(f"  compiler.cc: {'✓' if 'cc' in saved_config.get('compiler', {}) else '✗ MISSING'}")
        print(f"  compiler.flags: {'✓' if 'flags' in saved_config.get('compiler', {}) else '✗ MISSING'}")
        print(f"  compiler.debug: {'✗ REMOVED' if 'debug' not in saved_config.get('compiler', {}) else '✓ (should be removed)'}")
        print(f"  runtime.threads: {'✓' if 'threads' in saved_config.get('runtime', {}) else '✗ MISSING'}")
        print(f"  runtime.verbose: {'✓' if 'verbose' in saved_config.get('runtime', {}) else '✗ MISSING'}")

        # Verify the expected behavior
        assert 'compiler' in saved_config, "compiler section missing!"
        assert 'runtime' in saved_config, "runtime section missing!"
        assert 'debug' not in saved_config['compiler'], "debug key should be removed!"
        assert saved_config['compiler']['cc'] == 'gcc', "cc was lost!"
        assert saved_config['compiler']['flags'] == '-O2', "flags was lost!"
        assert saved_config['runtime']['threads'] == 4, "runtime.threads was lost!"
        assert saved_config['runtime']['verbose'] == True, "runtime.verbose was lost!"


def test_config_unset_entire_section():
    """Test removing an entire section."""
    from mccode_antlr.cli.management import config_unset

    with tempfile.TemporaryDirectory() as tmpdir:
        tmppath = Path(tmpdir)

        config_file = tmppath / 'config.yaml'
        initial_config = {
            'compiler': {
                'cc': 'gcc',
                'flags': '-O2'
            },
            'runtime': {
                'threads': 4
            }
        }
        with config_file.open('w') as f:
            yaml.dump(initial_config, f)

        print("\n" + "="*60)
        print("Test: Unsetting entire section")
        print("="*60)
        print("\nInitial config:")
        print(yaml.dump(initial_config))

        # Remove entire compiler section
        config_unset('compiler', str(tmppath))

        with config_file.open('r') as f:
            saved_config = yaml.safe_load(f)

        print("\nSaved config after unsetting compiler:")
        print(yaml.dump(saved_config))

        assert 'compiler' not in saved_config, "compiler section should be removed!"
        assert 'runtime' in saved_config, "runtime section was lost!"
        assert saved_config['runtime']['threads'] == 4, "runtime.threads was lost!"


if __name__ == '__main__':
    test_config_unset_preserves_keys()
    print("\n✅ Test 1 passed!")

    test_config_unset_entire_section()
    print("\n✅ Test 2 passed!")

    print("\n" + "="*60)
    print("All tests passed!")
    print("="*60)




"""Handles compilation of C translated instruments into usable binaries"""
from __future__ import annotations

from enum import Flag, auto
from typing import Union
from pathlib import Path
from mccode_antlr.instr import Instr
from mccode_antlr.translators.c import CTargetVisitor
from loguru import logger
from .check import compiled, gpu_only, mpi_only
from mccode_antlr import Flavor

class CBinaryTarget:
    class Type(Flag):
        single = auto()
        acc = auto()
        mpi = auto()
        nexus = auto()

    def __init__(self, mpi: bool = False, acc: bool = False, count: int = 1, nexus: bool = False):
        self.count = count
        if mpi and acc:
            self.type = CBinaryTarget.Type.acc | CBinaryTarget.Type.mpi
        elif mpi:
            self.type = CBinaryTarget.Type.mpi
        elif acc:
            self.type = CBinaryTarget.Type.acc
        else:
            self.type = CBinaryTarget.Type.single
        # completely orthogonal, probably a bad idea to include here at all:
        if nexus:
            self.type |= CBinaryTarget.Type.nexus

    def update(self, d: dict):
        self.count = d.get('count', self.count)
        d_mpi = d.get('mpi', self.type & CBinaryTarget.Type.mpi)
        d_acc = d.get('acc', self.type & CBinaryTarget.Type.acc)
        if d_mpi and d_acc:
            self.type = CBinaryTarget.Type.acc | CBinaryTarget.Type.mpi
        elif d_mpi:
            self.type = CBinaryTarget.Type.mpi
        elif d_acc:
            self.type = CBinaryTarget.Type.acc
        else:
            self.type = CBinaryTarget.Type.single
        if d.get('nexus', self.type & CBinaryTarget.Type.nexus):
            self.type |= CBinaryTarget.Type.nexus

    @property
    def compiler(self) -> str:
        from mccode_antlr.config import config
        if self.type & CBinaryTarget.Type.acc:
            # acc *or* acc + mpi uses the OpenACC compiler
            return config['acc'].as_str_expanded()
        elif self.type & CBinaryTarget.Type.mpi:
            # mpi (no-acc) uses the OpenMPI compiler
            return config['mpi']['cc'].as_str_expanded()
        # otherwise use the standard C compiler
        return config['cc'].as_str_expanded()

    @property
    def flags(self) -> list[str]:
        from mccode_antlr.config import config
        if self.type & CBinaryTarget.Type.acc:
            # acc *or* acc + mpi uses the OpenACC compiler *and* flags
            return config['flags']['acc'].as_str_expanded().split()
        # otherwise use the standard compiler flags
        return config['flags']['cc'].as_str_expanded().split()

    @property
    def linker_flags(self) -> list[str]:
        from mccode_antlr.config import config
        return config['flags']['ld'].as_str_expanded().split()

    @property
    def extra_flags(self) -> list[str]:
        """Only a little confusing ..."""
        from mccode_antlr.config import config
        extras = []
        if self.type & CBinaryTarget.Type.mpi:
            extras.extend(config['flags']['mpi'].as_str_expanded().split())
        if self.type & CBinaryTarget.Type.nexus:
            extras.extend(config['flags']['nexus'].as_str_expanded().split())
        return extras


def instrument_source(instrument: Instr, flavor: Flavor, config: dict, verbose: bool = None):
    if verbose is None:
        verbose = config.get('verbose', False)
    # TargetVisitor to uses instrument.registries (and the CTargetVisitor *appends* its LIBC_REGISTRY if necessary)
    visitor = CTargetVisitor(instrument, flavor=flavor, config=config, verbose=verbose)
    # Does the conversion by 'visiting' the instrument, then returns the full string of the generated source code
    return visitor.contents()


def linux_split_flags(instrument: Instr, target: CBinaryTarget):
    # the type of binary requested determines (some of) the required flags:
    compiler_flags = target.flags + target.extra_flags
    linker_flags = target.linker_flags
    # the instrument-defined flags are always(?) linker flags:
    # the flags in an instrument *might* contain ENV, CMD, GETPATH directives which need to be expanded via decode:
    linker_flags.extend(
        [word for flag in instrument.decoded_flags() for word in flag.split()])

    # Why is this addition necessary?
    if any('OPENACC' in word for word in compiler_flags) and any(
            'NeXus' in word for word in compiler_flags):
        compiler_flags.append('-D__GNUC__')
    return compiler_flags, linker_flags


def windows_split_flags(instrument: Instr, target: CBinaryTarget):
    # the type of binary requested determines (some of) the required flags:
    compiler_flags = target.flags + target.extra_flags
    linker_flags = target.linker_flags
    # instrument-defined flags may be for the compiler or linker.
    # cl.exe accepts a _wide_ array of flags, most of which we can't support
    # so, instead we are forced to handle things piecemeal.
    cl_flags = {
        'D': {'not': '/DYNAMICBASE',},
        'U': {},
        'p': {'has': '='},
        'std': {'has': '=', 'replace': ('=', ':')}
    }
    for flag in instrument.decoded_flags():
        flag = flag.strip()
        if not flag.startswith('/') and not flag.startswith('-'):
            # Not an actual flag? Punt for now
            linker_flags.append(flag)
        elif any(flag[1:].startswith(key) for key in cl_flags):
            d = next(d for key, d in cl_flags.items() if flag[1:].startswith(key))
            if ('not' in d and flag == d['not']) or ('has' in d and not d['has'] in flag):
                linker_flags.append(flag)
            else:
                if 'replace' in d:
                    flag = flag.replace(d['replace'][0], d['replace'][1])
                compiler_flags.append(flag)
        elif flag[1:].lower().startswith('l') or flag.lower().endswith('.lib'):
            linker_flags.append(flag)
        else:
            # Second punt; no match for our special (post / or -) strings
            linker_flags.append(flag)

    # Check for OpenACC or NeXus in flags?
    return compiler_flags, linker_flags


def linux_compile(compiler, compiler_flags, target, linker_flags, source):
    from subprocess import run
    # The solitary '-' specifies *where* the stdin source should be processed, which is critical for getting
    # linking flags right on (some) Linux systems
    command = [compiler, *compiler_flags, '-o', str(target), '-', *linker_flags]
    result = run(command, input=source, text=True, capture_output=True)
    return command, result


def windows_compile(compiler, compiler_flags, target, linker_flags, source):
    from subprocess import run
    parent = target.parent
    if not parent.is_dir():
        parent.mkdir(parents=True)
    write_to = target.with_suffix('.c')
    with write_to.open('w') as file:
        file.writelines(source)
    if '/link' not in linker_flags:
        linker_flags = ['/link'] + linker_flags
    # Specifying either of
    #   obj = [] if parent == Path() else [f'/Fo"{parent}\\"']
    # as *obj, would move generated object files next to the executable, but breaks
    # some aspect of compilation when done from Python (but not the command line)
    # FIXME Re-evaluate the need/advisability for no compiler warnings
    command = [compiler, *compiler_flags, str(write_to), '/W0', *linker_flags, f'/out:{target}']
    result = run(command, capture_output=True)
    return command, result


def _compile_instrument(
        instrument: Instr,
        target: CBinaryTarget,
        output: Union[str, Path] = None,
        replace: bool = False,
        dump_source: bool = False,
        source_file: str | None = None,
        **kwargs
):
    """Do the actual compilation -- should not be called directly by users

    Note
    ----
    If you are a user of the mccode-antlr module, call the `compile_instrument`
    gateway method instead to enable a cached check that your system compiler is
    configured correctly.
    """
    from os import R_OK, access
    from subprocess import run, CalledProcessError
    from platform import system
    from mccode_antlr.config import config
    logger.info(f'Compile {instrument.name}')
    # determine a name and location for the binary file
    if output is None:
        # use the current working directory if nothing provided
        output = Path()
    if not isinstance(output, Path):
        # allow for a string input to be interpreted as a path
        output = Path(output)
    if output.is_dir():
        # allow for the user to specify only the output *directory*
        output = output.joinpath(instrument.name).with_suffix(config['ext'].get(str))

    if output.exists() and not replace:
        return output

    source = instrument_source(instrument, **kwargs)
    if source_file or ('Windows' != system() and dump_source):
        source_file = source_file or Path().joinpath(output.parts[-1]).with_suffix('.c')
        logger.info(f'Source written in {source_file}')
        with open(source_file, 'w') as cfile:
            cfile.write(source)

    _compile = windows_compile if 'Windows' == system() else linux_compile
    _handle_flags = windows_split_flags if 'Windows' == system() else linux_split_flags
    compiler_flags, linker_flags = _handle_flags(instrument, target)
    command, result = _compile(target.compiler, compiler_flags, output, linker_flags, source)

    if result.returncode:
        stdout = result.stdout.decode() if isinstance(result.stdout, bytes) else result.stdout
        stderr = result.stderr.decode() if isinstance(result.stderr, bytes) else result.stderr
        raise RuntimeError(f"Compilation\n{' '.join(command)}\nfailed with output\n{stdout}\nand error\n{stderr}")
    if not output.exists():
        raise RuntimeError(f"Compilation should have produced {output}, but it does not appear to exist")
    if not access(output, R_OK):
        raise RuntimeError(f"{output} exists but is not an executable")
    return output


@gpu_only
def compile_acc_instrument(*args, **kwargs):
    return _compile_instrument(*args, **kwargs)


@mpi_only
def compile_mpi_instrument(*args, **kwargs):
    return _compile_instrument(*args, **kwargs)


@compiled
def compile_c_instrument(*args, **kwargs):
    return _compile_instrument(*args, **kwargs)


def compile_instrument(
        instrument: Instr,
        target: CBinaryTarget,
        output: Union[str, Path] = None,
        replace: bool = False,
        dump_source: bool = False,
        **kwargs
):
    """Compile an Instr object to one of the possible C targets

    Parameters
    ----------
    instrument: Instr
        The Instr to turn into a compiled binary
    target: CBinaryTarget
        The type of binary to produce: C99, OpenACC, MPI, etc.
    output:
        The path (and optionally name) where to store the produced binary
    replace:
        If true compilation will proceed even if a same-named path exists already
    dump_source:
        A diagnostic mode that outputs the generated C code into the current working
        directory
    """
    if target.type & CBinaryTarget.Type.acc:
        return compile_acc_instrument(instrument, target, output, replace, dump_source, **kwargs)
    if target.type & CBinaryTarget.Type.mpi:
        return compile_mpi_instrument(instrument, target, output, replace, dump_source, **kwargs)
    return compile_c_instrument(instrument, target, output, replace, dump_source, **kwargs)


def run_compiled_instrument(binary: Path, target: CBinaryTarget, options: str, capture=False, dry_run: bool = False):
    from subprocess import run, CalledProcessError
    from platform import system
    from mccode_antlr.config import config

    # If NeXus output is requested and the InstrumentDescriptionFile is needed, run a different script entirely...
    #   TODO think about actually doing this?
    # if target.flags & CBinaryTarget.Type.nexus and idf_required:
    #     run([config['idfgen'].get(str), str(binary), *options])

    command = []
    if target.type & CBinaryTarget.Type.mpi:
        # is the available MPI OpenMPI, mpich, or something else?
        res = run(['mpirun', '--version'], capture_output=True)
        if res.returncode != 0:
            raise RuntimeError(f"mpirun is not installed, MPI compilation failed")
        is_open_mpi = 'Open MPI' in res.stdout.decode()

        # we execute mpirun
        command.append(config['mpi']['run'].as_str_expanded())
        # which takes optional flags
        if target.count == 0 and not is_open_mpi:
            print("Using system default number of mpirun processes")
        else:
            # Even if zero, OpenMPI needs this to avoid interpreting options flags
            # for the binary as if they were for it.
            command.extend(['-np', str(target.count)])
        if config['machinefile'].exists():
            # --machinefile is only an OpenMPI option. mpich uses -f?
            machinefile = config['machinefile'].as_str_expanded()
            command.extend(['--machinefile' if is_open_mpi else '-f', machinefile])
        # and requires trickery if we want to restrict the GPU used
        if target.type & CBinaryTarget.Type.acc and 'Windows' != system():
            # Each worker should have the environment variable CUDA_VISIBLE_DEVICES defined as the value of
            # OMPI_COMM_WORLD_LOCAL_RANK, which gets sent by MPI to the worker. This assigment must take place
            # *on* the worker, which requires hijacking the executable that MPI runs
            command.append('acc_gpu_bind')
            raise NotImplementedError('CUDA GPU binding not yet implemented')
        if is_open_mpi:
            # mpich, at least, does not accept '--' between options and binary
            command.extend(['--'])

    binary = binary.resolve()
    if not binary.exists():
        raise RuntimeError(f'Can not execute {binary} since it does not exist.')

    # In normal operation, the binary is provided with options
    command.extend([str(binary), *options.split()])
    # which we then execute:
    if dry_run:
        logger.info(f'Would execute {command}')
        return ""
    result = run(command, capture_output=capture)
    if result.returncode and capture:
        raise RuntimeError(f'Execution of {command} failed with output\n{result.stdout}\n and error\n{result.stderr}')
    elif result.returncode:
        raise RuntimeError(f'Execution of {command} failed, see above for error message(s)')

    return (result.stdout + result.stderr) if capture else ""



import pytest
from mccode_antlr.run.range import MRange, EList, Singular, parse_scan_parameters


class TestParseScanParameters:
    def test_empty_list(self):
        """Test parsing an empty list returns an empty dictionary."""
        result = parse_scan_parameters([])
        assert result == {}

    def test_single_range_with_equals(self):
        """Test parsing a single range parameter with equals sign."""
        result = parse_scan_parameters(['param=1:0.5:3'])
        assert 'param' in result
        assert isinstance(result['param'], MRange)
        assert result['param'].start == 1
        assert result['param'].stop == 3
        assert result['param'].step == 0.5

    def test_single_range_without_step(self):
        """Test parsing a range without step (defaults to 1)."""
        result = parse_scan_parameters(['param=1:5'])
        assert 'param' in result
        assert isinstance(result['param'], MRange)
        assert result['param'].start == 1
        assert result['param'].stop == 5
        assert result['param'].step == 1

    def test_single_singular_with_equals(self):
        """Test parsing a single singular parameter with equals sign."""
        result = parse_scan_parameters(['param=42'])
        assert 'param' in result
        assert isinstance(result['param'], Singular)
        assert result['param'].value == 42
        assert result['param'].maximum == 1  # max_length is 1 for single singular

    def test_single_singular_space_separated(self):
        """Test parsing a single singular parameter with space separation."""
        result = parse_scan_parameters(['param', '42'])
        assert 'param' in result
        assert isinstance(result['param'], Singular)
        assert result['param'].value == 42

    def test_single_range_space_separated(self):
        """Test parsing a single range parameter with space separation."""
        result = parse_scan_parameters(['param', '1:0.5:3'])
        assert 'param' in result
        assert isinstance(result['param'], MRange)
        assert result['param'].start == 1
        assert result['param'].stop == 3
        assert result['param'].step == 0.5

    def test_multiple_ranges(self):
        """Test parsing multiple range parameters."""
        result = parse_scan_parameters(['a=1:5', 'b=2:0.5:4'])
        assert len(result) == 2
        assert isinstance(result['a'], MRange)
        assert isinstance(result['b'], MRange)
        assert result['a'].start == 1
        assert result['a'].stop == 5
        assert result['b'].start == 2
        assert result['b'].stop == 4
        assert result['b'].step == 0.5

    def test_singular_with_range_updates_maximum(self):
        """Test that Singular maximum is updated to match range length."""
        result = parse_scan_parameters(['a=1:5', 'b=10'])
        assert isinstance(result['a'], MRange)
        assert isinstance(result['b'], Singular)
        # The range a=1:5 has 5 elements (1, 2, 3, 4, 5)
        expected_max = len(result['a'])
        assert result['b'].maximum == expected_max

    def test_mixed_parameters(self):
        """Test parsing a mix of ranges and singular values."""
        result = parse_scan_parameters(['x=0:10', 'y', '5', 'z=1:2:9'])
        assert len(result) == 3
        assert 'x' in result
        assert 'y' in result
        assert 'z' in result

    def test_float_values(self):
        """Test parsing float values."""
        result = parse_scan_parameters(['param=1.5:0.25:3.5'])
        assert result['param'].start == 1.5
        assert result['param'].stop == 3.5
        assert result['param'].step == 0.25

    def test_negative_values(self):
        """Test parsing negative values in ranges."""
        result = parse_scan_parameters(['param=-5:1:5'])
        assert result['param'].start == -5
        assert result['param'].stop == 5
        assert result['param'].step == 1

    def test_singular_string_value(self):
        """Test parsing a singular parameter with a string value."""
        result = parse_scan_parameters(['param=myfile'])
        assert isinstance(result['param'], Singular)
        assert result['param'].value == 'myfile'

    def test_invalid_parameter_raises_error(self):
        """Test that an invalid parameter format raises ValueError."""
        with pytest.raises(ValueError, match='Invalid parameter'):
            parse_scan_parameters(['invalid'])

    def test_preserves_parameter_case(self):
        """Test that parameter names preserve case."""
        result = parse_scan_parameters(['MyParam=10', 'UPPERCASE=20'])
        assert 'MyParam' in result
        assert 'UPPERCASE' in result

    def test_multiple_singulars_all_get_maximum_one(self):
        """Test that multiple singulars without ranges get maximum 1."""
        result = parse_scan_parameters(['a=10', 'b=20'])
        assert isinstance(result['a'], Singular)
        assert isinstance(result['b'], Singular)
        assert result['a'].maximum == 1
        assert result['b'].maximum == 1

    def test_singular_maximum_matches_longest_range(self):
        """Test that singulars get maximum from the longest range."""
        # a has 5 elements (1,2,3,4,5), b has 3 elements (1,3,5)
        result = parse_scan_parameters(['a=1:5', 'b=1:2:5', 'c=42'])
        assert isinstance(result['c'], Singular)
        max_len = max(len(result['a']), len(result['b']))
        assert result['c'].maximum == max_len


class TestEList:
    """Tests for explicit list (EList) parsing functionality."""

    def test_elist_from_str_integers(self):
        """Test EList.from_str with integer values."""
        result = EList.from_str('1,2,3,4,5')
        assert result.values == [1, 2, 3, 4, 5]
        assert all(isinstance(v, int) for v in result.values)

    def test_elist_from_str_floats(self):
        """Test EList.from_str with float values."""
        result = EList.from_str('1.5,2.5,3.5')
        assert result.values == [1.5, 2.5, 3.5]
        assert all(isinstance(v, float) for v in result.values)

    def test_elist_from_str_mixed_int_float(self):
        """Test EList.from_str with mixed integer and float values."""
        result = EList.from_str('1,2.5,3,4.5')
        assert result.values == [1, 2.5, 3, 4.5]
        assert isinstance(result.values[0], int)
        assert isinstance(result.values[1], float)

    def test_elist_from_str_single_value(self):
        """Test EList.from_str with a single value."""
        result = EList.from_str('42')
        assert result.values == [42]
        assert len(result) == 1

    def test_elist_from_str_negative_values(self):
        """Test EList.from_str with negative values."""
        result = EList.from_str('-1,-2.5,3,-4')
        assert result.values == [-1, -2.5, 3, -4]

    def test_elist_len(self):
        """Test EList __len__ method."""
        result = EList.from_str('1,2,3')
        assert len(result) == 3

    def test_elist_iter(self):
        """Test EList __iter__ method."""
        result = EList.from_str('10,20,30')
        values = list(result)
        assert values == [10, 20, 30]

    def test_elist_getitem(self):
        """Test EList __getitem__ method."""
        result = EList.from_str('10,20,30')
        assert result[0] == 10
        assert result[1] == 20
        assert result[2] == 30

    def test_elist_getitem_out_of_range(self):
        """Test EList __getitem__ raises IndexError for out-of-range index."""
        result = EList.from_str('1,2,3')
        with pytest.raises(IndexError, match='Index 5 out of range'):
            _ = result[5]

    def test_elist_getitem_negative_index(self):
        """Test EList __getitem__ raises IndexError for negative index."""
        result = EList.from_str('1,2,3')
        with pytest.raises(IndexError, match='Index -1 out of range'):
            _ = result[-1]

    def test_elist_str(self):
        """Test EList __str__ method."""
        result = EList.from_str('1,2,3')
        assert str(result) == '1,2,3'

    def test_elist_repr(self):
        """Test EList __repr__ method."""
        result = EList.from_str('1,2,3')
        assert repr(result) == 'EList(1,2,3)'

    def test_elist_equality(self):
        """Test EList __eq__ method."""
        list1 = EList.from_str('1,2,3')
        list2 = EList.from_str('1,2,3')
        list3 = EList.from_str('1,2,4')
        assert list1 == list2
        assert not (list1 == list3)

    def test_elist_in_parse_scan_parameters(self):
        """Test EList parsing via parse_scan_parameters."""
        result = parse_scan_parameters(['values=1,2,3,4'])
        assert 'values' in result
        assert isinstance(result['values'], EList)
        assert result['values'].values == [1, 2, 3, 4]

    def test_elist_space_separated_parsing(self):
        """Test EList parsing with space separation."""
        result = parse_scan_parameters(['values', '1,2,3'])
        assert 'values' in result
        assert isinstance(result['values'], EList)
        assert result['values'].values == [1, 2, 3]

    def test_elist_with_range_maximum_not_updated(self):
        """Test that EList values are not affected by range maximum."""
        result = parse_scan_parameters(['a=1:10', 'b=5,10,15'])
        assert isinstance(result['a'], MRange)
        assert isinstance(result['b'], EList)
        # EList values should remain unchanged
        assert result['b'].values == [5, 10, 15]
        assert len(result['b']) == 3

    def test_multiple_elists(self):
        """Test parsing multiple EList parameters."""
        result = parse_scan_parameters(['x=1,2,3', 'y=4,5,6'])
        assert isinstance(result['x'], EList)
        assert isinstance(result['y'], EList)
        assert result['x'].values == [1, 2, 3]
        assert result['y'].values == [4, 5, 6]

    def test_elist_with_large_precision_floats(self):
        """Test EList with high precision float values."""
        result = EList.from_str('0.123456789,0.987654321')
        assert result.values[0] == 0.123456789
        assert result.values[1] == 0.987654321

    def test_elist_scientific_notation(self):
        """Test EList with scientific notation values."""
        result = EList.from_str('1e-3,2.5e2,3e10')
        assert result.values == [1e-3, 2.5e2, 3e10]

    def test_elist_preserves_zero(self):
        """Test EList correctly handles zero values."""
        result = EList.from_str('0,1,0,2')
        assert result.values == [0, 1, 0, 2]
        assert result.values[0] == 0
        assert result.values[2] == 0

    def test_elist_with_singular_and_range(self):
        """Test combining EList with Singular and MRange."""
        result = parse_scan_parameters(['a=1:3', 'b=10', 'c=1.1,2.2,3.3'])
        assert isinstance(result['a'], MRange)
        assert isinstance(result['b'], Singular)
        assert isinstance(result['c'], EList)
        assert len(result['a']) == 3
        assert result['b'].value == 10
        assert result['c'].values == [1.1, 2.2, 3.3]

    def test_elist_direct_constructor(self):
        """Test EList direct construction with a list."""
        result = EList([1, 2, 3])
        assert result.values == [1, 2, 3]
        assert len(result) == 3

    def test_elist_direct_constructor_empty(self):
        """Test EList direct construction with an empty list."""
        result = EList([])
        assert result.values == []
        assert len(result) == 0

    def test_elist_long_list(self):
        """Test EList with a long list of values."""
        values_str = ','.join(str(i) for i in range(100))
        result = EList.from_str(values_str)
        assert len(result) == 100
        assert result.values == list(range(100))

    def test_elist_very_small_floats(self):
        """Test EList with very small float values."""
        result = EList.from_str('1e-10,1e-20,1e-30')
        assert result.values == [1e-10, 1e-20, 1e-30]

    def test_elist_very_large_floats(self):
        """Test EList with very large float values."""
        result = EList.from_str('1e10,1e20,1e30')
        assert result.values == [1e10, 1e20, 1e30]

    def test_elist_positive_explicit_sign(self):
        """Test EList with explicit positive signs."""
        result = EList.from_str('+1,+2.5,+3')
        assert result.values == [1, 2.5, 3]

    def test_elist_mixed_signs(self):
        """Test EList with mixed positive and negative values."""
        result = EList.from_str('-1,+2,-3.5,+4.5')
        assert result.values == [-1, 2, -3.5, 4.5]

    def test_elist_equality_different_lengths(self):
        """Test EList equality with different lengths raises error."""
        list1 = EList.from_str('1,2,3')
        list2 = EList.from_str('1,2')
        with pytest.raises(ValueError):
            _ = list1 == list2

    def test_elist_iteration_multiple_times(self):
        """Test that EList can be iterated multiple times."""
        result = EList.from_str('1,2,3')
        first_pass = list(result)
        second_pass = list(result)
        assert first_pass == second_pass == [1, 2, 3]

    def test_elist_sum_of_values(self):
        """Test that EList works with built-in sum function."""
        result = EList.from_str('1,2,3,4,5')
        assert sum(result) == 15

    def test_elist_in_list_comprehension(self):
        """Test EList in list comprehension."""
        result = EList.from_str('1,2,3')
        doubled = [v * 2 for v in result]
        assert doubled == [2, 4, 6]

    def test_elist_float_str_roundtrip(self):
        """Test that float values survive str conversion roundtrip."""
        original = EList.from_str('1.5,2.5,3.5')
        roundtrip = EList.from_str(str(original))
        assert original == roundtrip

    def test_elist_int_str_roundtrip(self):
        """Test that integer values survive str conversion roundtrip."""
        original = EList.from_str('1,2,3')
        roundtrip = EList.from_str(str(original))
        assert original == roundtrip

    def test_elist_two_values(self):
        """Test EList with exactly two values."""
        result = EList.from_str('10,20')
        assert result.values == [10, 20]
        assert len(result) == 2

    def test_elist_getitem_last_element(self):
        """Test EList __getitem__ with index of last element."""
        result = EList.from_str('10,20,30')
        assert result[2] == 30

    def test_elist_negative_scientific_notation(self):
        """Test EList with negative values in scientific notation."""
        result = EList.from_str('-1e-3,-2.5e2,-3e10')
        assert result.values == [-1e-3, -2.5e2, -3e10]



from http.client import UnimplementedFileMode

from loguru import logger

from ..grammar import McCompParser as Parser, McCompVisitor
from .comp import Comp
from ..common import ComponentParameter, Expr, MetaData
from ..common.visitor import add_common_visitors
from ..grammar.McCompParser import McCompParser


class CompVisitor(McCompVisitor):
    def __init__(self, parent, filename, instance_name=None):
        self.parent = parent  # the instrument (handler?) that wanted to read this component
        self.filename = filename
        self.state = Comp()
        self.instance_name = instance_name

    def visitProg(self, ctx: Parser.ProgContext):
        self.state = Comp()
        self.visit(ctx.component_definition())
        return self.state

    def visitComponentDefineNew(self, ctx: Parser.ComponentDefineNewContext):
        self.state.name = str(ctx.Identifier())
        if ctx.NoAcc() is not None:
            self.state.no_acc()
        self.visitChildren(ctx)  # Use the visitor methods to fill in details of the state

    def visitComponentDefineCopy(self, ctx: Parser.ComponentDefineCopyContext):
        from copy import deepcopy
        new_name, copy_from = [str(x) for x in ctx.Identifier()]
        if self.parent is None:
            raise RuntimeError("Can not copy a component definition without a parent instrument")
        copy_from_comp = self.parent.get_component(copy_from)
        # pull from that component ... deepcopy _just_ in case
        self.state = deepcopy(copy_from_comp)
        # update the name to match what we're going to call this component
        self.state.name = new_name
        # add parameters, overwrite all provided details
        if ctx.NoAcc() is not None:
            self.state.no_acc()
        self.visitChildren(ctx)  # Use the visitor methods to overwrite details of the state

    def visitCategory(self, ctx: Parser.CategoryContext):
        # Return the provided identifier or string literal, minus quotes
        self.state.category = str(ctx.StringLiteral())[1:-1] if ctx.Identifier() is None else str(ctx.Identifer())

    def visitComponent_define_parameters(self, ctx: Parser.Component_define_parametersContext):
        for parameter in self.visit(ctx.component_parameters()):
            self.state.add_define(parameter)

    def visitComponent_set_parameters(self, ctx: Parser.Component_set_parametersContext):
        for parameter in self.visit(ctx.component_parameters()):
            self.state.add_setting(parameter)

    def visitComponent_out_parameters(self, ctx: Parser.Component_out_parametersContext):
        for parameter in self.visit(ctx.component_parameters()):
            self.state.add_output(parameter)

    def visitComponent_parameters(self, ctx: Parser.Component_parametersContext):
        return [self.visit(x) for x in ctx.component_parameter()]

    def visitComponentParameterDouble(self, ctx: Parser.ComponentParameterDoubleContext):
        name = str(ctx.Identifier())
        value = None
        if ctx.Assign() is not None:
            # protect against a literal '0' provided ... which doesn't match IntegerLiteral for some reason
            value = 0 if ctx.expr() is None else self.visit(ctx.expr())
        return ComponentParameter(name=name, value=Expr.float(value))

    def visitComponentParameterInteger(self, ctx: Parser.ComponentParameterIntegerContext):
        name = str(ctx.Identifier())
        value = None
        if ctx.Assign() is not None:
            # protect against a literal '0' provided ... which doesn't match IntegerLiteral for some reason
            value = 0 if ctx.expr() is None else self.visit(ctx.expr())
        return ComponentParameter(name=name, value=Expr.int(value))

    def visitComponentParameterString(self, ctx: Parser.ComponentParameterStringContext):
        name = str(ctx.Identifier())
        default = None
        if ctx.Assign() is not None:
            default = 'NULL' if ctx.StringLiteral() is None else str(ctx.StringLiteral())
        return ComponentParameter(name=name, value=Expr.str(default))

    def visitComponentParameterVector(self, ctx: Parser.ComponentParameterVectorContext):
        from ..common import Value, DataType, ShapeType
        name = str(ctx.Identifier(0))
        if ctx.Assign() is not None and ctx.initializerlist() is not None:
            value = self.visit(ctx.initializerlist())
            value.data_type = DataType.float
        else:
            default = None
            if ctx.Assign() is not None:
                default = "NULL"
                if ctx.Identifier(1) is not None:
                    default = str(ctx.Identifier(1))
            value = Expr(Value(default, DataType.float, _shape=ShapeType.vector))
        return ComponentParameter(name=name, value=value)

    def visitComponentParameterSymbol(self, ctx: Parser.ComponentParameterSymbolContext):
        raise RuntimeError("McCode symbol parameter type not supported yet")

    def visitComponentParameterDoubleArray(self, ctx: Parser.ComponentParameterDoubleArrayContext):
        # 'vector' is really just an alias for 'double *', right?
        return self.visitComponentParameterVector(ctx)

    def visitComponentParameterIntegerArray(self, ctx: Parser.ComponentParameterIntegerArrayContext):
        from ..common import Value, DataType, ShapeType
        name = str(ctx.Identifier(0))
        if ctx.assign() is not None and ctx.initializerlist() is not None:
            value = self.visit(ctx.initializerlist())
            value.data_type = DataType.int
        else:
            default = None
            if ctx.assign() is not None:
                default = "NULL"
                if ctx.Identifier(1) is not None:
                    default = str(ctx.Identifier(1))
            value = Expr(Value(default, DataType.int, _shape=ShapeType.vector))
        return ComponentParameter(name=name, value=value)

    def visitDependency(self, ctx: Parser.DependencyContext):
        if ctx.StringLiteral() is not None:
            # the flags are the literal string without its quotes:
            self.parent.add_c_flags(str(ctx.StringLiteral()).strip('"'))

    def visitComponent_trace(self, ctx: McCompParser.Component_traceContext):
        self.state.TRACE(*self.multi_block("trace", ctx.multi_block()))

    def visitDeclare(self, ctx:McCompParser.DeclareContext):
        self.state.DECLARE(*self.multi_block("declare", ctx.multi_block()))

    def visitShare(self, ctx:McCompParser.ShareContext):
        self.state.SHARE(*self.multi_block("share", ctx.multi_block()))

    def visitInitialise(self, ctx:McCompParser.InitialiseContext):
        self.state.INITIALIZE(*self.multi_block("initialize", ctx.multi_block()))

    def visitUservars(self, ctx: Parser.UservarsContext):
        self.state.USERVARS(*self.multi_block("user", ctx.multi_block()))

    def visitSave(self, ctx:McCompParser.SaveContext):
        self.state.SAVE(*self.multi_block("save", ctx.multi_block()))

    def visitFinally(self, ctx:McCompParser.FinallyContext):
        self.state.FINALLY(*self.multi_block("final", ctx.multi_block()))

    def visitDisplay(self, ctx:McCompParser.DisplayContext):
        self.state.DISPLAY(*self.multi_block("display", ctx.multi_block()))

    def visitMetadata(self, ctx: Parser.MetadataContext):
        filename, line_number, metadata = self.visit(ctx.unparsed_block())
        mime = ctx.mime.text if ctx.mime.type == Parser.Identifier else ctx.mime.text[1:-1]
        name = ctx.name.text if ctx.name.type == Parser.Identifier else ctx.name.text[1:-1]
        metadata = MetaData.from_component_tokens(source=self.state.name, mimetype=mime, name=name, value=metadata)
        self.state.add_metadata(metadata)

    def visitUnparsed_block(self, ctx: Parser.Unparsed_blockContext):
        # We want to extract the source-file line number (and filename) for use in the C-preprocessor
        # via `#line {number} "{filename}"` directives, for more expressive error handling
        line_number = None if ctx.start is None else ctx.start.line
        content = str(ctx.UnparsedBlock())[2:-2]
        return self.filename, line_number, content

    # FIXME There *are* no statements in McCode, so all identifiers always produce un-parsable values.
    def visitAssignment(self, ctx: Parser.AssignmentContext):
        line_number = None if ctx.start is None else ctx.start.line
        raise RuntimeError(f"{self.filename}: {line_number} -- assignment statements are not (yet) supported")

    def visitExpressionPrevious(self, ctx: Parser.ExpressionPreviousContext):
        # The very-special no-good expression use of PREVIOUS where it is replaced by the last component's name
        raise RuntimeError('PREVIOUS is not a valid expression in Comp definitions')

    def visitExpressionMyself(self, ctx: Parser.ExpressionMyselfContext):
        # The even-worse expression use of MYSELF to refer to the current being-constructed component's name
        return Expr.str(self.instance.name)

    def multi_block(self, part: str, ctx: Parser.Multi_blockContext):
        """Common visitor for {part} unparsed_block? ((INHERIT identifier)|(EXTEND unparsed_block))*

        Ensures that the correct 'part' is pulled from named component definition(s)
        and that the definitions and new unparsed blocks are inserted in their given
        order.
        """
        blocks = dict()
        for ident in ctx.Identifier():
            comp = self.parent.get_component(str(ident))
            blocks[ident.getSourceInterval()[0]] = getattr(comp, part)
        for unparsed in ctx.unparsed_block():
            blocks[unparsed.getSourceInterval()[0]] = (self.visit(unparsed),)
        return [b for n in sorted(blocks.keys()) for b in blocks[n]]



add_common_visitors(CompVisitor)